@@ -14,22 +14,27 @@ import shutil
 import socket
 import subprocess
 import sys
+import tempfile
 import time
 import traceback
 from collections import OrderedDict
+from contextlib import contextmanager
 
+import docker
+import termcolor
 import yaml
+from dt_shell.constants import DTShellConstants
+from dt_shell.env_checks import check_executable_exists, InvalidEnvironment, check_docker_environment
+from dt_shell.remote import ConnectionError, make_server_request, DEFAULT_DTSERVER
 from duckietown_challenges.challenge import EvaluationParameters, SUBMISSION_CONTAINER_TAG
-from duckietown_challenges.challenge_results import read_challenge_results, ChallengeResults, ChallengeResultsStatus
+from duckietown_challenges.challenge_results import read_challenge_results, ChallengeResults, ChallengeResultsStatus, \
+    NoResultsFound
 from duckietown_challenges.constants import CHALLENGE_SOLUTION_OUTPUT_DIR, CHALLENGE_RESULTS_DIR, \
     CHALLENGE_DESCRIPTION_DIR, \
     CHALLENGE_EVALUATION_OUTPUT_DIR, ENV_CHALLENGE_NAME, ENV_CHALLENGE_STEP_NAME, CHALLENGE_PREVIOUS_STEPS_DIR
 from duckietown_challenges.utils import safe_yaml_dump, friendly_size, indent
 from six import BytesIO
 
-from dt_shell.constants import DTShellConstants
-from dt_shell.env_checks import check_executable_exists, InvalidEnvironment, check_docker_environment
-from dt_shell.remote import ConnectionError, make_server_request, DEFAULT_DTSERVER
 from . import __version__
 from .runner_cache import copy_to_cache, get_file_from_cache
 
@@ -68,6 +73,8 @@ def get_token_from_shell_config():
 def dt_challenges_evaluator():
     try:
         dt_challenges_evaluator_()
+    except SystemExit:  # --help
+        raise
     except BaseException as e:
         msg = traceback.format_exc(e)
         elogger.error(msg)
@@ -143,27 +150,37 @@ Usage:
     args = dict(do_upload=do_upload, do_pull=do_pull, more_features=more_features,
                 delete=delete, evaluator_name=evaluator_name, machine_id=machine_id,
                 tmpdir=tmpdir, token=token)
+    ndots = 0
+
     if parsed.continuous:
 
         timeout = 5.0  # seconds
         multiplier = 1.0
-        max_multiplier = 10
+        multiplier_grow = 1.5
+        max_multiplier = 5
         while True:
             multiplier = min(multiplier, max_multiplier)
+            t0 = time.time()
             try:
                 go_(None, reset=False, **args)
                 multiplier = 1.0
             except NothingLeft:
-                sys.stderr.write('.')
+                delta = time.time() - t0
+
+                sys.stderr.write('No work for me yet. Server answers in %.1f seconds.\n' % delta)
+                ndots += 1
+                if ndots == 5:
+                    # sys.stderr.write(' no work for me yet.\n')
+                    ndots = 0
                 # time.sleep(timeout * multiplier)
                 # elogger.info('No submissions available to evaluate.')
             except ConnectionError as e:
                 elogger.error(e)
-                multiplier *= 1.5
+                multiplier *= multiplier_grow
             except BaseException as e:
                 msg = 'Uncaught exception:\n%s' % traceback.format_exc(e)
                 elogger.error(msg)
-                multiplier *= 1.5
+                multiplier *= multiplier_grow
 
             time.sleep(timeout * multiplier)
 
@@ -239,6 +256,60 @@ class DockerComposeFail(Exception):
     pass
 
 
+def get_FORMAT_datefmt():
+    pre = '%(asctime)s|%(name)s|%(filename)s:%(lineno)s|%(funcName)s(): '
+    pre = termcolor.colored(pre, attrs=['dark'])
+    FORMAT = pre + "%(message)s"
+    datefmt = "%H:%M:%S"
+    return FORMAT, datefmt
+
+
+@contextmanager
+def setup_logging(wd):
+    old_stderr = sys.stderr
+    old_stdout = sys.stdout
+    if not os.path.exists(wd):
+        os.makedirs(wd)
+    my_stderr = Tee(os.path.join(wd, 'stderr.log'), sys.stderr)
+    sys.stderr = my_stderr
+    my_stdout = Tee(os.path.join(wd, 'stdout.log'), sys.stdout)
+    sys.stdout = my_stdout
+    ch = logging.StreamHandler(my_stderr.file)
+    ch.setLevel(logging.DEBUG)
+
+    root = logging.getLogger()
+    FORMAT, datefmt = get_FORMAT_datefmt()
+    formatter = logging.Formatter(FORMAT, datefmt=datefmt)
+    ch.setFormatter(formatter)
+    root.addHandler(ch)
+
+    try:
+        yield
+    finally:
+
+        my_stdout.flush()
+        my_stderr.flush()
+
+        for s in [my_stderr, my_stdout]:
+            try:
+                fn = os.path.splitext(s.name)[0] + '.html'
+                with open(fn, 'w') as f:
+                    from ansi2html import Ansi2HTMLConverter
+                    conv = Ansi2HTMLConverter()
+                    v = s.getvalue().strip()
+                    if not v:
+                        v = '(No output was produced.)'
+                    html = conv.convert(v)
+                    f.write(html)
+                os.remove(s.name)
+            except:
+                pass
+
+        sys.stderr = old_stderr
+        sys.stdout = old_stdout
+        root.removeHandler(ch)
+
+
 def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluator_name, machine_id, tmpdir,
         token=None):
     features = get_features(more_features)
@@ -251,56 +322,57 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
     evaluator_version = 'DC:%s;DCR:%s;DTS:%s' % (version_dcr, version_dc, version_dts)
     process_id = evaluator_name
 
+    timeout_server = 60
     res = dtserver_work_submission(token, submission_id, machine_id, process_id, evaluator_version,
-                                   features=features, reset=reset)
+                                   features=features, reset=reset, timeout=timeout_server)
 
     if 'job_id' not in res:
         msg = 'Could not find jobs: %s' % res['msg']
         raise NothingLeft(msg)
 
     job_id = res['job_id']
+    elogger.info(safe_yaml_dump(res))
 
+    challenge_name = res['challenge_name']
+    challenge_step_name = res['step_name']
+    submission_id = res['submission_id']
+
+    elogger.info('Evaluating job %s' % job_id)
+
+    uploaded = []
     try:
-        elogger.info(safe_yaml_dump(res))
 
-        # if res['protocol'] != 'p1':
-        #     msg = 'invalid protocol %s' % res['protocol']
-        #     elogger.error(msg)
-        #     raise Exception(msg)
+        steps2artefacts = res['steps2artefacts']
 
-        challenge_name = res['challenge_name']
-        challenge_step_name = res['step_name']
-        submission_id = res['submission_id']
+        solution_container = res['parameters']['hash']
 
-        elogger.info('Evaluating job %s' % job_id)
+        wd = os.path.join(tmpdir, challenge_name, 'submission%d' % submission_id,
+                          '%s-%s-job%s' % (challenge_step_name, evaluator_name, job_id))
+
+        if not os.path.exists(wd):
+            os.makedirs(wd)
+
+        own_logs_dir = os.path.join(wd, 'logs', 'challenges-runner')
 
         aws_config = res['aws_config']
         if aws_config and do_upload:
             try_s3(aws_config)
 
-            # evaluation_protocol = challenge_parameters['protocol']
-        # assert evaluation_protocol == 'p1'
+        try:
 
-        # you get this from the server
+            with setup_logging(own_logs_dir):
+                challenge_parameters_ = EvaluationParameters.from_yaml(res['challenge_parameters'])
 
-        # from rpath to Artefact.as_dict(
-        steps2artefacts = res['steps2artefacts']
+                project = 'job%s-%s' % (job_id, random.randint(1, 10000))
+                if not do_upload:
+                    aws_config = None
 
-        # for k, v in steps2artefacts_.items():
-        #     steps2artefacts[k] = Artefact.from_yaml()
-        solution_container = res['parameters']['hash']
-        challenge_parameters_ = EvaluationParameters.from_yaml(res['challenge_parameters'])
-
-        wd = os.path.join(tmpdir, challenge_name, 'submission%d' % submission_id,
-                          '%s-%s-job%s' % (challenge_step_name, evaluator_name, job_id))
-        project = 'job%s-%s' % (job_id, random.randint(1, 10000))
-        if not do_upload:
-            aws_config = None
-
-        cr = run_single(wd, aws_config, steps2artefacts, challenge_parameters_, solution_container, challenge_name,
-                        challenge_step_name, project, do_pull)
-
-        uploaded = upload_files(wd, aws_config)
+                cr = run_single(wd, aws_config, steps2artefacts, challenge_parameters_, solution_container,
+                                challenge_name,
+                                challenge_step_name, project, do_pull)
+        finally:
+            if do_upload:
+                uploaded = upload_files(wd, aws_config)
 
         if delete:
             cmd = ['down']
@@ -308,14 +380,22 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
 
         if delete:
             shutil.rmtree(wd)
-    except BaseException as e:  # XXX
-        msg = 'Uncaught exception:\n%s' % traceback.format_exc(e)
+
+    except KeyboardInterrupt:
+        msg = 'KeyboardInterrupt:\n%s' % traceback.format_exc()
+        elogger.error(msg)
+        status = ChallengeResultsStatus.ABORTED
+        cr = ChallengeResults(status, msg, scores={})
+
+    except BaseException:
+        msg = 'Uncaught exception:\n%s' % traceback.format_exc()
         elogger.error(msg)
         status = ChallengeResultsStatus.ERROR
         cr = ChallengeResults(status, msg, scores={})
-        uploaded = []
 
-    msg = 'This is what is being reported.\n\nstatus = %s\n\n%s' % (cr.get_status(), cr.msg)
+    msg = ('Reporting status %s for job %s step %s submission %s.\n\n%s\n\n%s' %
+           (cr.get_status(), job_id, challenge_step_name, submission_id,
+            indent(cr.msg, ' msg |'), indent(json.dumps(cr.get_stats(), indent=4), 'stats |')))
     if cr.get_status() != ChallengeResultsStatus.SUCCESS:
         elogger.error(msg)
     else:
@@ -334,7 +414,8 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
                                 machine_id=machine_id,
                                 process_id=process_id,
                                 evaluator_version=evaluator_version,
-                                uploaded=uploaded)
+                                uploaded=uploaded,
+                                timeout=timeout_server)
             break
         except BaseException as e:
             msg = 'Could not report: %s' % e
@@ -347,6 +428,12 @@ def go_(submission_id, do_pull, more_features, do_upload, delete, reset, evaluat
 def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_container, challenge_name,
                challenge_step_name, project, do_pull):
     prepare_dir(wd, aws_config, steps2artefacts)
+    tomonitor = []
+    for service_name, service_def in challenge_parameters.services.items():
+        if service_def.image == SUBMISSION_CONTAINER_TAG:
+            continue
+        else:
+            tomonitor.append(service_name)
     config = get_config(challenge_parameters, solution_container, challenge_name, challenge_step_name)
     config_yaml = yaml.safe_dump(config, encoding='utf-8', indent=4, allow_unicode=True)
     # elogger.debug('YAML:\n' + config_yaml)
@@ -375,7 +462,9 @@ def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_c
         valid_config = False
 
     if valid_config:
-        cr = run(wd, project, do_pull)
+        services_names = list(challenge_parameters.services)
+        # print('services: %s' % services_names)
+        cr = run_one(wd, project, do_pull, services=services_names, monitor=tomonitor[0])
 
         write_logs(wd, project, services=config['services'])
     else:
@@ -385,17 +474,81 @@ def run_single(wd, aws_config, steps2artefacts, challenge_parameters, solution_c
     return cr
 
 
-def run(wd, project, do_pull):
+def get_id_for_service(wd, project, service_name):
+    cmd = ['ps', '-q', service_name]
+
+    try:
+        o = run_docker(wd, project, cmd, get_output=True)
+        container_id = o.strip()  # \n at the end
+    except DockerComposeFail:
+        raise
+    return container_id
+
+
+class EvaluatorTimeout(Exception):
+    pass
+
+
+def write_logs_worker(service_name, container_id):
+    elogger.info('writing logs for %s ' % container_id)
+
+    client = check_docker_environment()
+    from dateutil import parser
+    container = client.containers.get(container_id)
+    for x in container.logs(stream=True, stdout=True, stderr=True,
+                            timestamps=True):
+        i = x.index(' ')
+        timestamp = x[:i]
+        message = x[i + 1:].rstrip()  # \n
+        ts = parser.parse(timestamp)
+        s = "%s %s %s" % (ts.strftime('%H:%M:%S'), service_name, message)
+        print(s)
+        if container.status != 'running':
+            break
+
+    exit_code = exit_code_for_container(container_id)
+
+    elogger.info('container %s terminated with %s' % (service_name, exit_code))
+
+
+def exit_code_for_container(container_id):
+    cmd = ['docker', 'inspect', container_id, "--format={{.State.ExitCode}}"]
+    out = subprocess.check_output(cmd)
+    return int(out.strip())
+
+
+class Workers(object):
+    workers = []
+
+
+def stream_logs(service_name, container_id):
+    from multiprocessing import Process
+    p = Process(target=write_logs_worker, args=(service_name, container_id,))
+    p.start()
+    Workers.workers.append(p)
+
+
+def teminate_workers():
+    for p in list(Workers.workers):
+        elogger.debug('terminating %s' % p)
+        Workers.workers.remove(p)
+        p.terminate()
+    elogger.debug('terminated all workers.')
+
+DEFAULT_TIMEOUT = 1800
+
+def run_one(wd, project, do_pull, monitor, services, timeout=DEFAULT_TIMEOUT):
     client = check_docker_environment()
 
     try:
         if do_pull:
             elogger.info('pulling containers')
-            cmd = ['pull']
+            cmd = ['pull',  # '--ignore-pull-failures'
+                   ]
             run_docker(wd, project, cmd)
 
         try:
-            pruned = client.networks.prune()
+            pruned = client.networks.prune(filters=dict(until='1h'))
             elogger.debug('pruned: %s' % pruned)
         except BaseException as e:  # XXX not critical
             pass
@@ -404,20 +557,88 @@ def run(wd, project, do_pull):
         # cmd = ['create', '--force-recreate']
         # run_docker(wd, project, cmd)
 
-        elogger.info('Running containers')
+        elogger.info('Running containers %s' % services)
         cmd = ['up',
+               '-d',
                # '--remove-orphans',
-               '--abort-on-container-exit'
+               # '--abort-on-container-exit'
                ]
         run_docker(wd, project, cmd)
 
-        cr = read_challenge_results(wd)
+        service2id = {}
+        for service_name in services:
+            service2id[service_name] = get_id_for_service(wd, project, service_name)
+            stream_logs(service_name, service2id[service_name])
+        # container = client.containers.get(container_id)
 
-    except BaseException as e:  # XXX
-        msg = 'Uncaught exception while running Docker Compose:\n%s' % traceback.format_exc(e)
+        # last_notice = time.time()
+        t0 = time.time()
+
+        client = check_docker_environment()
+        while True:
+            exited = []
+            should_break = False
+            for service_name, container_id in service2id.items():
+                container = client.containers.get(container_id)
+                msg = 'service %s: %s' % (service_name, container.status)
+                elogger.debug(msg)
+                if container.status != 'running':
+                    exited.append(service_name)
+                    exit_code = exit_code_for_container(container_id)
+                    if exit_code:
+                        msg = '%s exited with code %s' % (service_name, exit_code)
+                        elogger.error(msg)
+                        should_break = True
+
+            if should_break:
+                break
+
+            if monitor in exited:
+                break
+
+            delta = time.time() - t0
+            if delta > timeout:
+                msg = 'Waited %s for container to finish. Giving up. ' % delta
+                raise EvaluatorTimeout(msg)
+
+            time.sleep(4)
+
+        try:
+            cr = read_challenge_results(wd)
+        except NoResultsFound:
+            msg = '''\
+The result file is not found. This usually means that the evaluator did not finish
+and some times that there was an import error.
+
+Check the evaluator log to see what happened.'''
+            elogger.error(msg)
+            status = ChallengeResultsStatus.ERROR
+            cr = ChallengeResults(status, msg, scores={})
+
+    except EvaluatorTimeout as e:
+        msg = 'Timeout:\n\n%s' % e
         elogger.error(msg)
         status = ChallengeResultsStatus.ERROR
         cr = ChallengeResults(status, msg, scores={})
+    except DockerComposeFail as e:
+        msg = 'Error while running Docker Compose:\n\n%s' % e
+        elogger.error(msg)
+        status = ChallengeResultsStatus.ERROR
+        cr = ChallengeResults(status, msg, scores={})
+
+    except KeyboardInterrupt:
+        msg = 'KeyboardInterrupt:\n%s' % traceback.format_exc()
+        elogger.error(msg)
+        status = ChallengeResultsStatus.ABORTED
+        cr = ChallengeResults(status, msg, scores={})
+
+    except BaseException:  # XXX
+        msg = 'Uncaught exception while running Docker Compose:\n%s' % traceback.format_exc()
+        elogger.error(msg)
+        status = ChallengeResultsStatus.ERROR
+        cr = ChallengeResults(status, msg, scores={})
+    finally:
+        teminate_workers()
 
     return cr
 
@@ -443,6 +664,9 @@ def prepare_dir(wd, aws_config, steps2artefacts):
 
 
 def get_config(challenge_parameters_, solution_container, challenge_name, challenge_step_name):
+    client = docker.from_env()
+    from docker.errors import NotFound
+    # Adding the submission container
     for service_def in challenge_parameters_.services.values():
         service_def.build = None
 
@@ -451,9 +675,36 @@ def get_config(challenge_parameters_, solution_container, challenge_name, challe
 
     config = challenge_parameters_.as_dict()
 
-    # Adding the submission container
+    do_trick = False
     for service in config['services'].values():
         image_digest = service.pop('image_digest', None)
+
+        if do_trick:
+            if image_digest:
+                # If we have the image then we don't bother with pulling the tag
+                try:
+                    c = client.images.get(image_digest)
+                except NotFound:
+                    msg = 'Could not find the image for %s by digest %s; will need to pull.' % (
+                        service['image'], image_digest)
+                    elogger.info(msg)
+                else:
+                    msg = 'I already have the exact image for %s' % service['image']
+                    elogger.info(msg)
+                    #
+                    #
+                    if True:
+
+                        n = len('91bac885982d')
+                        partial_id = image_digest.replace('sha256:', '')[:n]
+                        service['image'] = partial_id
+                    else:
+                        image = service['image']
+                        # remove tag
+                        if ':' in image:
+                            image = image.split(':')[0]
+                        service['image'] = image + '@' + image_digest
+
         service.pop('build', None)
 
         # if service['image'] == SUBMISSION_CONTAINER_TAG:
@@ -511,13 +762,14 @@ def write_logs(wd, project, services):
         os.makedirs(logdir)
 
     for service in services:
-        cmd = ['ps', '-q', service]
+        container_id = get_id_for_service(wd, project, service)
 
-        try:
-            o = run_docker(wd, project, cmd, get_output=True)
-            container_id = o.strip()  # \n at the end
-        except DockerComposeFail:
-            continue
+        # cmd = ['ps', '-q', service]
+        # try:
+        #     o = run_docker(wd, project, cmd, get_output=True)
+        #     container_id = o.strip()  # \n at the end
+        # except DockerComposeFail:
+        #     continue
 
         if not container_id:
             logs = 'Service "%s" was not started.' % service
@@ -525,22 +777,85 @@ def write_logs(wd, project, services):
         else:
             elogger.info('Found container ID = %r' % container_id)
 
-            logs = logs_for_container(client, container_id)
+            for fname, options in [
+                ('combined', dict(stderr=True, stdout=True)),
+                ('stderr', dict(stderr=True, stdout=False)),
+                ('stdout', dict(stdout=True, stderr=False)),
+            ]:
 
-        from ansi2html import Ansi2HTMLConverter
-        conv = Ansi2HTMLConverter()
-        html = conv.convert(logs)
+                logs = logs_for_container(client, container_id, **options)
 
-        logdir_service = os.path.join(logdir, service)
-        if not os.path.exists(logdir_service):
-            os.makedirs(logdir_service)
-        fn = os.path.join(logdir_service, 'log.txt')
-        with open(fn, 'w') as f:
-            f.write(logs)
+                from ansi2html import Ansi2HTMLConverter
+                conv = Ansi2HTMLConverter()
+                html = conv.convert(logs)
 
-        fn = os.path.join(logdir_service, 'log.html')
-        with open(fn, 'w') as f:
-            f.write(html)
+                logdir_service = os.path.join(logdir, service)
+                if not os.path.exists(logdir_service):
+                    os.makedirs(logdir_service)
+                fn = os.path.join(logdir_service, '%s.log' % fname)
+                with open(fn, 'w') as f:
+                    f.write(logs)
+
+                fn = os.path.join(logdir_service, '%s.html' % fname)
+                with open(fn, 'w') as f:
+                    f.write(html)
+
+
+# class Tee1(object):
+#     def __init__(self, other):
+#         self.file = BytesIO()
+#         self.other = other
+#         # self.stdout = sys.stdout
+#
+#     def __del__(self):
+#         # sys.stdout = self.stdout
+#         self.file.close()
+#
+#     def write(self, data):
+#         self.file.write(data)
+#         self.other.write(data)
+#         self.other.flush()
+#
+#     def flush(self):
+#         self.other.flush()
+#         self.file.flush()
+#
+#     def getvalue(self):
+#         return self.file.getvalue()
+#
+#     def fileno(self):
+#         return self.other.fileno()
+
+
+class Tee(object):
+    def __init__(self, name, f):
+        self.file = open(name, 'wb')
+        self.name = name
+        self.original = f
+
+    def close(self):
+        if hasattr(self, 'file'):
+            if self.file is not None:
+                self.file.close()
+                self.file = None
+
+    def __del__(self):
+        self.close()
+
+    def write(self, data):
+        self.file.write(data)
+        self.original.write(data)
+
+    def flush(self):
+        self.file.flush()
+        self.original.flush()
+
+    def fileno(self):
+        return self.file.fileno()
+
+    def getvalue(self):
+        with open(self.name) as f:
+            return f.read()
 
 
 def run_docker(cwd, project, cmd0, get_output=False):
@@ -548,23 +863,42 @@ def run_docker(cwd, project, cmd0, get_output=False):
     # elogger.info('Running:\n\t%s' % " ".join(cmd0) + '\n\n in %s' % cwd)
     elogger.debug('Running: %s' % " ".join(cmd0))
 
+    d = tempfile.mkdtemp()
+
+    fn1 = os.path.join(d, 'docker-stdout.txt')
+    fn2 = os.path.join(d, 'docker-stderr.txt')
+    elogger.debug('Saving stdout to %s' % fn1)
+
+
+    tee_stdout = Tee(fn1, sys.stdout)
+    tee_stderr = Tee(fn2, sys.stderr)
+
+
     try:
         if get_output:
             return subprocess.check_output(cmd0, cwd=cwd, stderr=sys.stderr)
         else:
-            subprocess.check_call(cmd0, cwd=cwd, stdout=sys.stdout, stderr=sys.stderr)
+            subprocess.check_call(cmd0, cwd=cwd, stdout=tee_stdout, stderr=tee_stderr)
     except subprocess.CalledProcessError as e:
         msg = 'Could not run %s:\n\n %s' % (cmd0, indent(e, '  >  '))
-        msg += '\n\n%s' % indent(e.output, ' docker-compose stdout  | ')
-        # msg += '\n\n%s' % indent(e., ' docker-compose stderr  | ')
+        msg += '\n\n%s' % indent(tee_stdout.getvalue(), 'stdout | ')
+        msg += '\n\n%s' % indent(tee_stderr.getvalue(), 'stderr | ')
         raise DockerComposeFail(msg)
+    finally:
+        try:
+            os.unlink(fn1)
+            os.unlink(fn2)
+        except:
+            pass
+        pass
+
 
 
 def upload_files(wd, aws_config, ignore_patterns=('.DS_Store',)):
     toupload = get_files_to_upload(wd, ignore_patterns=ignore_patterns)
 
     if not aws_config:
-        msg = 'Not uploading artefacts because AWS config not passed.'
+        msg = 'Not uploading artifacts because AWS config not passed.'
         elogger.info(msg)
         uploaded = only_copy_to_cache(toupload)
     else:
@@ -672,11 +1006,9 @@ def get_files_to_upload(path, ignore_patterns=()):
     return toupload
 
 
-def logs_for_container(client, container_id):
-    logs = ''
+def logs_for_container(client, container_id, stdout, stderr):
     container = client.containers.get(container_id)
-    for c in container.logs(stdout=True, stderr=True, stream=True, timestamps=True):
-        logs += c
+    logs = container.logs(stdout=stdout, stderr=stderr, timestamps=True)
     return logs
 
 
@@ -775,7 +1107,7 @@ def compute_sha256hex(filename):
 
 
 def dtserver_report_job(token, job_id, result, stats, machine_id,
-                        process_id, evaluator_version, uploaded):
+                        process_id, evaluator_version, uploaded, timeout):
     endpoint = '/take-submission'
     method = 'POST'
     data = {'job_id': job_id,
@@ -787,12 +1119,12 @@ def dtserver_report_job(token, job_id, result, stats, machine_id,
             'uploaded': uploaded
             }
     try:
-        return make_server_request(token, endpoint, data=data, method=method, suppress_user_msg=True)
+        return make_server_request(token, endpoint, data=data, method=method, timeout=timeout, suppress_user_msg=True)
     except:
-        return make_server_request(token, endpoint, data=data, method=method)
+        return make_server_request(token, endpoint, data=data, method=method, timeout=timeout)
 
 
-def dtserver_work_submission(token, submission_id, machine_id, process_id, evaluator_version, features, reset):
+def dtserver_work_submission(token, submission_id, machine_id, process_id, evaluator_version, features, reset, timeout):
     endpoint = '/take-submission'
     method = 'GET'
     data = {'submission_id': submission_id,
@@ -801,10 +1133,10 @@ def dtserver_work_submission(token, submission_id, machine_id, process_id, evalu
             'evaluator_version': evaluator_version,
             'features': features,
             'reset': reset}
-    try:
-        return make_server_request(token, endpoint, data=data, method=method, suppress_user_msg=True)
-    except:
-        return make_server_request(token, endpoint, data=data, method=method)
+    # try:
+    return make_server_request(token, endpoint, data=data, method=method, timeout=timeout, suppress_user_msg=True)
+    # except:
+    #     return make_server_request(token, endpoint, data=data, method=method, timeout=timeout)
 
 
 def create_index_files(wd, job_id):
